/***************************
 * dicom sort by Tony Voet *
 ***************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <dicom.h>

static int	sort_info(const char *);
static void	sort_target(const char *);
static int	sort_move(const char *);
static void	sort_free(void);
static int	sort_mkdir(const char *);

static struct
{
  char	**patient,**id,
	**date,**modality,**bodypart,**coil,
	**sequence,**orientation,**contrast;
  int	image;
}
info;

#ifndef FILENAME_MAX
#define FILENAME_MAX 1024
#endif

static char target[4*FILENAME_MAX];

/********
 * main *
 ********/

int main(int argc,char *argv[])
{
  int i;

  if (argc<3)
  {
    fprintf(stderr,"%s (%s)\n",*argv,dicom_version);
    fprintf(stderr,"Usage: %s data-directory dicom-file ...\n",*argv);

    return -1;
  }

  dicom_log_name(*argv);

  umask(-1-(S_IRWXU|S_IRWXG));

  for (i=2; i<argc; i++)
  {
    if (!sort_info(argv[i]))
    {
      sort_target(argv[1]);
      sort_move(argv[i]);
    }
    sort_free();
  }

  return 0;
}

/********
 * info *
 ********/

static int sort_info(const char *file)
{
  ELEMENT *e;

  dicom_log(INFO,"sort_info()");

  if (dicom_open(file))
    return -1;

  memset(&info,0,sizeof(info));
  info.image=-1;

  for (;;)
  {
    e=dicom_element();
    if (!e)
      break;

    if (e->group==0x0008)
    {
      if (e->element==0x0020)
      {
        if (dicom_load(DA))
          return -2;
        dicom_clean();
        info.date=e->value.DA;
        continue;
      }

      if (e->element==0x0060)
      {
        if (dicom_load(CS))
          return -3;
        dicom_clean();
        info.modality=e->value.CS;
        continue;
      }
    }

    if (e->group==0x0010)
    {
      if (e->element==0x0010)
      {
        if (dicom_load(PN))
          return -4;
        dicom_clean();
        info.patient=e->value.PN;
        continue;
      }

      if (e->element==0x0020)
      {
        if (dicom_load(LO))
          return -5;
        dicom_clean();
        info.id=e->value.LO;
        continue;
      }
    }

    if (e->group==0x0018)
    {
      if (e->element==0x0010)
      {
        if (dicom_load(LO))
          return -6;
        dicom_clean();
        info.contrast=e->value.LO;
        continue;
      }

      if (e->element==0x0015)
      {
        if (dicom_load(CS))
          return -7;
        dicom_clean();
        info.bodypart=e->value.CS;
        continue;
      }

      if (e->element==0x0024)
      {
        if (dicom_load(SH))
          return -8;
        dicom_clean();
        info.sequence=e->value.SH;
        continue;
      }

      if (e->element==0x1250)
      {
        if (dicom_load(SH))
          return -9;
        dicom_clean();
        info.coil=e->value.SH;
        continue;
      }
    }

    if (e->group==0x0020)
    {
      if (e->element==0x0013)
      {
        if (dicom_load(IS))
          return -10;
        dicom_clean();
        info.image=atoi(*e->value.IS);
        free(e->value.IS);
        continue;
      }

      if (e->element==0x0037)
      {
        if (dicom_load(DS))
          return -11;

        if (e->vm!=6 && e->vm)
          dicom_log(WARNING,"Wrong Image Orientation VM");
        else
        {
          dicom_clean();
          info.orientation=e->value.DS;
        }

        dicom_close();
        break;
      }
    }

    if (dicom_skip())
      return -12;
  }

  return 0;
}

/**********
 * target *
 **********/

static void sort_target(const char *root)
{
  char *last;

  dicom_log(INFO,"sort_target()");

  strcpy(target,root);
  last=target+strlen(target)-1;
  if (*last=='/')
    *last=0;

  sort_mkdir(target);
  strcat(target,"/");

  if (info.patient)
    strcat(target,*info.patient);
  else
    strcat(target,"Unknown");

  if (info.id)
  {
    strcat(target,"-");
    strcat(target,*info.id);
  }

  sort_mkdir(target);
  strcat(target,"/");

  if (info.date)
  {
    strcat(target,*info.date);
    strcat(target,"-");
  }

  if (info.modality)
  {
    strcat(target,*info.modality);
    strcat(target,"-");
  }

  if (info.bodypart)
  {
    strcat(target,*info.bodypart);
    strcat(target,"-");
  }
  else
    if (info.coil)
    {
      strcat(target,*info.coil);
      strcat(target,"-");
    }

  if (info.sequence)
  {
    strcat(target,*info.sequence);
    strcat(target,"-");
  }

  if (info.orientation)
  {
    if (atof(info.orientation[2])==0.0)
      if (atof(info.orientation[5])==0.0)
        strcat(target,"Tra-");

    if (atof(info.orientation[1])==0.0)
      if (atof(info.orientation[4])==0.0)
        strcat(target,"Cor-");

    if (atof(info.orientation[0])==0.0)
      if (atof(info.orientation[3])==0.0)
        strcat(target,"Sag-");
  }

  if (info.contrast)
    if (strcmp(*info.contrast,"NONE"))
    {
      strcat(target,*info.contrast);
      strcat(target,"-");
    }

  target[strlen(target)-1]=0;

  sort_mkdir(target);
}

/********
 * move *
 ********/

static int sort_move(const char *file)
{
  DIR		*dir;
  struct dirent	*entry;
  struct stat	tmp;
  int		new=-1,i;

  dicom_log(INFO,"sort_move()");

  if (info.image>0)
  {
    sprintf(target+strlen(target),"/%.4u.dcm",info.image);

    if (stat(target,&tmp))
      new=0;
    else
      *strrchr(target,'/')=0;
  }

  if (new)
  {
    dir=opendir(target);
    if (!dir)
    {
      dicom_log(ERROR,"Unable to open directory");
      return -1;
    }

    for(;;)
    {
      entry=readdir(dir);
      if (!entry)
	break;

      i=atoi(entry->d_name);
      if (i>new)
	new=i;
    }

    if (closedir(dir))
      dicom_log(WARNING,"Unable to close directory");

    sprintf(target+strlen(target),"/%.4u.dcm",new+1);
  }

  dicom_log(NOTICE,target);

  if (link(file,target))
  {
    dicom_log(ERROR,"Unable to link file");
    return -2;
  }

  if (unlink(file))
    dicom_log(WARNING,"Unable to unlink file");

  return 0;
}

/********
 * free *
 ********/

static void sort_free(void)
{
  dicom_log(INFO,"sort_free()");

  free(info.patient);
  free(info.id);

  free(info.date);
  free(info.modality);
  free(info.bodypart);
  free(info.coil);
  free(info.sequence);
  free(info.orientation);
  free(info.contrast);
}

/*********
 * mkdir *
 *********/

static int sort_mkdir(const char *dir)
{
  struct stat	tmp;
  int		fail, i=0;

  dicom_log(INFO,"sort_mkdir()");

  /* change spaces into underscores */
  while(dir[i] != '\0') {
    if (dir[i] == ' ') dir[i]='_';
    i+=1;
  }

  if (!stat(dir,&tmp))
    return 0;

  dicom_log(NOTICE,dir);

  fail=mkdir(dir,S_IRWXU|S_IRWXG);

  if (fail)
    dicom_log(ERROR,"Unable to create directory");

  return fail;
}
